'use strict';

var _getConfigs = function () {
  var _ref = _asyncToGenerator(function* (configFile, files) {
    var isPathIgnored = void 0;
    var getConfigForFile = void 0;

    if (eslint.ESLint) {
      var esLint = new eslint.ESLint({
        // Ignore any config applicable depending on the location on the filesystem
        useEslintrc: false,
        // Point to the particular config
        overrideConfigFile: configFile
      });
      isPathIgnored = esLint.isPathIgnored.bind(esLint);
      getConfigForFile = esLint.calculateConfigForFile.bind(esLint);
    } else {
      var cliEngine = new eslint.CLIEngine({
        // Ignore any config applicable depending on the location on the filesystem
        useEslintrc: false,
        // Point to the particular config
        configFile
      });
      isPathIgnored = cliEngine.isPathIgnored.bind(cliEngine);
      getConfigForFile = cliEngine.getConfigForFile.bind(cliEngine);
    }

    var configs = files.map(function () {
      var _ref2 = _asyncToGenerator(function* (filePath) {
        return (yield isPathIgnored(filePath)) ? false : getConfigForFile(filePath);
      });

      return function (_x3) {
        return _ref2.apply(this, arguments);
      };
    }());
    return new Set((yield Promise.all(configs)).filter(Boolean));
  });

  return function _getConfigs(_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();

var _getConfig = function () {
  var _ref3 = _asyncToGenerator(function* (configFile, files) {
    return Array.from((yield _getConfigs(configFile, files))).reduce(function (prev, item) {
      return Object.assign(prev, item, { rules: Object.assign({}, prev.rules, item.rules) });
    }, {});
  });

  return function _getConfig(_x4, _x5) {
    return _ref3.apply(this, arguments);
  };
}();

var createRuleFinder = function () {
  var _ref5 = _asyncToGenerator(function* (specifiedFile, options) {
    var configFile = _getConfigFile(specifiedFile);

    var _options$ext = options.ext,
        ext = _options$ext === undefined ? ['.js'] : _options$ext;

    var extensionRegExp = _createExtensionRegExp(ext);
    var files = glob.sync(`**/*`, { dot: true, matchBase: true }).filter(function (file) {
      return extensionRegExp.test(file);
    });

    var config = yield _getConfig(configFile, files);

    return new RuleFinder(config, options);
  });

  return function createRuleFinder(_x6, _x7) {
    return _ref5.apply(this, arguments);
  };
}();

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

var path = require('path');

var eslint = require('eslint');
var glob = require('glob');
var difference = require('./array-diff');
var getSortedRules = require('./sort-rules');
var normalizePluginName = require('./normalize-plugin-name');

function _getConfigFile(specifiedFile) {
  if (specifiedFile) {
    if (path.isAbsolute(specifiedFile)) {
      return specifiedFile;
    }
    return path.join(process.cwd(), specifiedFile);
  }
  // This is not being called with an arg. Use the package.json `main`
  return require(path.join(process.cwd(), 'package.json')).main;
}

function _getCurrentNamesRules(config) {
  return config.rules ? Object.keys(config.rules) : [];
}

function _isDeprecated(rule) {
  return rule && rule.meta && rule.meta.deprecated;
}

function _notDeprecated(rule) {
  return !_isDeprecated(rule);
}

function _getPluginRules(config) {
  var pluginRules = new Map();
  var plugins = config.plugins;
  /* istanbul ignore else */
  if (plugins) {
    plugins.forEach(function (plugin) {
      var normalized = normalizePluginName(plugin);
      var pluginConfig = require(normalized.module);
      var rules = pluginConfig.rules === undefined ? {} : pluginConfig.rules;

      Object.keys(rules).forEach(function (ruleName) {
        return pluginRules.set(`${normalized.prefix}/${ruleName}`, rules[ruleName]);
      });
    });
  }
  return pluginRules;
}

function _getCoreRules() {
  return (eslint.Linter ? new eslint.Linter() : eslint.linter).getRules();
}

function _filterRuleNames(ruleNames, rules, predicate) {
  return ruleNames.filter(function (ruleName) {
    return predicate(rules.get(ruleName));
  });
}

function _isNotCore(rule) {
  return rule.indexOf('/') !== '-1';
}

function _escapeRegExp(str) {
  return str.replace(/[\\^$.*+?()[\]{}|]/g, '\\$&');
}

function _createExtensionRegExp(extensions) {
  var normalizedExts = extensions.map(function (ext) {
    return _escapeRegExp(ext.startsWith('.') ? ext.slice(1) : ext);
  });

  return new RegExp(`.\\.(?:${normalizedExts.join("|")})$`);
}

function RuleFinder(config, _ref4) {
  var omitCore = _ref4.omitCore,
      includeDeprecated = _ref4.includeDeprecated;

  var currentRuleNames = _getCurrentNamesRules(config);
  if (omitCore) {
    currentRuleNames = currentRuleNames.filter(_isNotCore);
  }

  var pluginRules = _getPluginRules(config); // eslint-disable-line vars-on-top
  var coreRules = _getCoreRules();
  var allRules = omitCore ? pluginRules : new Map([].concat(_toConsumableArray(coreRules), _toConsumableArray(pluginRules)));

  var allRuleNames = [].concat(_toConsumableArray(allRules.keys()));
  var pluginRuleNames = [].concat(_toConsumableArray(pluginRules.keys()));
  if (!includeDeprecated) {
    allRuleNames = _filterRuleNames(allRuleNames, allRules, _notDeprecated);
    pluginRuleNames = _filterRuleNames(pluginRuleNames, pluginRules, _notDeprecated);
  }
  var deprecatedRuleNames = _filterRuleNames(currentRuleNames, allRules, _isDeprecated);
  var dedupedRuleNames = [].concat(_toConsumableArray(new Set(allRuleNames)));
  var unusedRuleNames = difference(dedupedRuleNames, currentRuleNames);

  // Get all the current rules instead of referring the extended files or documentation
  this.getCurrentRules = function () {
    return getSortedRules(currentRuleNames);
  };

  // Get all the current rules' particular configuration
  this.getCurrentRulesDetailed = function () {
    return config.rules;
  };

  // Get all the plugin rules instead of referring the extended files or documentation
  this.getPluginRules = function () {
    return getSortedRules(pluginRuleNames);
  };

  // Get all the available rules instead of referring eslint and plugin packages or documentation
  this.getAllAvailableRules = function () {
    return getSortedRules(dedupedRuleNames);
  };

  this.getUnusedRules = function () {
    return getSortedRules(unusedRuleNames);
  };

  // Get all the current rules that are deprecated
  this.getDeprecatedRules = function () {
    return getSortedRules(deprecatedRuleNames);
  };
}

module.exports = function (specifiedFile) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  return createRuleFinder(specifiedFile, options);
};